.in-iframe .page-header {
  display: none;
}
.in-iframe .container {
  padding-top: 10px;
  width: 100%;
}
.iframe-only,
.in-iframe .iframe-hidden {
  display: none;
}
.in-iframe .iframe-only {
  display: block;
}
