{% extends "pretixbase/email/base.html" %}
{% load eventurl %}
{% load i18n %}
{% block content %}
    <tr>
        <td class="containertd">
            <div class="content">
                {{ body|safe }}
            </div>
        </td>
    </tr>
    {% if order %}
        <tr>
            <td class="gap"></td>
        </tr>
        <tr>
            <td class="order containertd">
                <div class="content">
                    {% trans "You are receiving this email because you placed an order for the following event:" %}<br>
                    <strong>{% trans "Event:" %}</strong> {{ event.name }}<br>
                    <strong>{% trans "Order code:" %}</strong> {{ order.code }}<br>
                    <strong>{% trans "Order date:" %}</strong> {{ order.datetime|date:"SHORT_DATE_FORMAT" }}<br>
                    <a href="{% abseventurl event "presale:event.order" order=order.code secret=order.secret %}">
                        {% trans "View order details" %}
                    </a>
                </div>
            </td>
        </tr>
    {% endif %}
	{% if signature %}
        <tr>
            <td class="gap"></td>
        </tr>
        <tr>
            <td class="order containertd">
                <div class="content">
                    {{ signature | safe }}
                </div>
            </td>
        </tr>
    {% endif %}
{% endblock %}
